"""Scrape data from Web."""

import json
import os
from pathlib import Path
from typing import List, Type, Tuple, Dict, Optional

import pandas as pd
from pydantic import BaseModel, create_model, ConfigDict
from rich.panel import Panel

from par_scrape.utils import console
from par_scrape.lib.llm_config import LlmConfig
from par_scrape.lib.llm_providers import LlmProvider


def save_raw_data(raw_data: str, run_name: str, output_folder: Path) -> str:
    """
    Save raw data to a file.

    Args:
        raw_data (str): The raw data to save.
        run_name (str): The run name to use in the filename.
        output_folder (str, optional): The folder to save the file in. Defaults to 'output'.

    Returns:
        str: The path to the saved file.
    """
    # Ensure the output folder exists
    os.makedirs(output_folder, exist_ok=True)

    # Save the raw markdown data with run_name in filename
    raw_output_path = os.path.join(output_folder, f"rawData_{run_name}.md")
    with open(raw_output_path, "wt", encoding="utf-8") as f:
        f.write(raw_data)
    console.print(
        Panel(f"Raw data saved to [bold green]{raw_output_path}[/bold green]")
    )
    return raw_output_path


def create_dynamic_listing_model(field_names: List[str]) -> Type[BaseModel]:
    """
    Dynamically creates a Pydantic model based on provided fields.

    Args:
        field_names (List[str]): A list of names of the fields to extract from the markdown.

    Returns:
        Type[BaseModel]: A dynamically created Pydantic model.
    """
    # Create field definitions using aliases for Field parameters
    field_definitions = {field: (str, ...) for field in field_names}
    # Dynamically create the model with all fields
    dynamic_listing_model = create_model(
        "DynamicListingModel",
        **field_definitions,  # type: ignore
    )  # type: ignore
    dynamic_listing_model.model_config = ConfigDict(arbitrary_types_allowed=True)
    return dynamic_listing_model


def create_listings_container_model(listing_model: Type[BaseModel]) -> Type[BaseModel]:
    """
    Create a container model that holds a list of the given listing model.

    Args:
        listing_model (Type[BaseModel]): The Pydantic model for individual listings.

    Returns:
        Type[BaseModel]: A container model for a list of listings.
    """
    return create_model("DynamicListingsContainer", listings=(List[listing_model], ...))


# pylint: disable=too-many-positional-arguments
def format_data(
    data: str,
    dynamic_listings_container: Type[BaseModel],
    model: str,
    ai_provider: LlmProvider,
    extraction_prompt: Optional[Path] = None,
    ai_base_url: Optional[str] = None,
) -> BaseModel:
    """
    Format data using the specified AI provider's API.

    Args:
        data (str): The input data to format.
        dynamic_listings_container (Type[BaseModel]): The Pydantic model to use for parsing.
        model (str): The AI model to use for processing.
        ai_provider (LlmProvider): The AI provider to use for processing.
        extraction_prompt (Path): Path to the extraction prompt file.
        ai_base_url (str): The base URL for the AI provider.

    Returns:
        BaseModel: The formatted data as a Pydantic model instance.
    """
    if not extraction_prompt:
        extraction_prompt = Path(__file__).parent / "extraction_prompt.md"
    try:
        system_message = extraction_prompt.read_text(encoding="utf-8")
    except FileNotFoundError:
        console.print(
            f"[bold red]Extraction prompt file not found: {extraction_prompt}[/bold red]"
        )
        raise

    user_message = f"Extract the following information from the provided text:\nPage content:\n\n{data}"

    try:
        llm_config = LlmConfig(
            provider=ai_provider, model_name=model, temperature=0, base_url=ai_base_url
        )
        chat_model = llm_config.build_chat_model()

        structure_model = chat_model.with_structured_output(
            dynamic_listings_container  # , include_raw=True
        )
        data = structure_model.invoke(
            [
                ("system", system_message),
                ("user", user_message),
            ]
        )  # type: ignore
        if isinstance(data, BaseModel):
            return data
        console.print(data)
        raise ValueError("Error in API call. Did not return a Pydantic BaseModel")
    except Exception as e:  # pylint: disable=broad-exception-caught
        console.print(
            f"[bold red]Error in API call or parsing response:[/bold red] {str(e)}"
        )
        return dynamic_listings_container(listings=[])


def save_formatted_data(
    formatted_data: BaseModel, run_name: str, output_folder: Path
) -> Tuple[pd.DataFrame | None, Dict[str, Path]]:
    """
    Save formatted data to JSON, Excel, CSV, and Markdown files.

    Args:
        formatted_data (BaseModel): The formatted data to save.
        run_name (str): The run name to use in the filenames.
        output_folder (str, optional): The folder to save the files in. Defaults to 'output'.

    Returns:
        Tuple[pd.DataFrame | None, Dict[str, str]]: The DataFrame created from the formatted data and a dictionary of
        file paths, or None and an empty dict if an error occurred.
    """
    file_paths: Dict[str, Path] = {}
    # Ensure the output folder exists
    os.makedirs(output_folder, exist_ok=True)

    # Prepare formatted data as a dictionary
    formatted_data_dict = formatted_data.model_dump()

    # Save the formatted data as JSON with run_name in filename
    json_output_path = output_folder / f"sorted_data_{run_name}.json"
    json_output_path.write_text(
        json.dumps(formatted_data_dict, indent=4), encoding="utf-8"
    )

    console.print(
        Panel(
            f"Formatted data saved to JSON at [bold green]{json_output_path}[/bold green]"
        )
    )
    file_paths["json"] = json_output_path

    # Prepare data for DataFrame
    if isinstance(formatted_data_dict, dict):
        # If the data is a dictionary containing lists, assume these lists are records
        data_for_df = (
            next(iter(formatted_data_dict.values()))
            if len(formatted_data_dict) == 1
            else formatted_data_dict
        )
    elif isinstance(formatted_data_dict, list):
        data_for_df = formatted_data_dict
    else:
        raise ValueError(
            "Formatted data is neither a dictionary nor a list, cannot convert to DataFrame"
        )

    # Create DataFrame
    try:
        df = pd.DataFrame(data_for_df)
        console.print(Panel("[bold green]DataFrame created successfully.[/bold green]"))

        # Save the DataFrame to an Excel file
        excel_output_path = output_folder / f"sorted_data_{run_name}.xlsx"
        df.to_excel(excel_output_path, index=False)
        console.print(
            Panel(
                f"Formatted data saved to Excel at [bold green]{excel_output_path}[/bold green]"
            )
        )
        file_paths["excel"] = excel_output_path

        # Save the DataFrame to a CSV file
        csv_output_path = output_folder / f"sorted_data_{run_name}.csv"
        df.to_csv(csv_output_path, index=False)
        console.print(
            Panel(
                f"Formatted data saved to CSV at [bold green]{csv_output_path}[/bold green]"
            )
        )
        file_paths["csv"] = csv_output_path

        # Save the DataFrame as a Markdown table
        markdown_output_path = output_folder / f"sorted_data_{run_name}.md"
        markdown_output_path.write_text(
            df.to_markdown(index=False) or "", encoding="utf-8"
        )
        console.print(
            Panel(
                f"Formatted data saved as Markdown table at [bold green]{markdown_output_path}[/bold green]"
            )
        )
        file_paths["md"] = markdown_output_path

        return df, file_paths
    except Exception as e:  # pylint: disable=broad-except
        console.print(
            f"[bold red]Error creating DataFrame or saving files:[/bold red] {str(e)}"
        )
        return None, {}



"""Utility functions for par_scrape."""

import tiktoken
from rich.console import Console

# Initialize rich console
console = Console()


def estimate_tokens(text: str) -> int:
    """
    Estimate the number of tokens in a given text.

    Args:
        text (str): The input text to estimate the number of tokens for.

    Returns:
        int: The estimated number of tokens.
    """
    try:
        return int(len(text) / 3) + 3
    except KeyError:
        return len(text)


def trim_to_token_limit(text: str, model: str, max_tokens: int = 200000) -> str:
    """
    Trim text to a specified token limit for a given model.

    Args:
        text (str): The input text to trim.
        model (str): The model name to use for tokenization.
        max_tokens (int, optional): The maximum number of tokens. Defaults to 200000.

    Returns:
        str: The trimmed text.
    """
    try:
        encoder = tiktoken.encoding_for_model(model)
        tokens = encoder.encode(text)
        if len(tokens) > max_tokens:
            return encoder.decode(tokens[:max_tokens])
    except KeyError:
        tokens = estimate_tokens(text)
        if tokens > max_tokens:
            return text[: max_tokens * 3]

    return text



"""LLM provider types."""

from __future__ import annotations

from enum import Enum


class LlmProvider(str, Enum):
    """Llm provider types."""

    OLLAMA = "Ollama"
    OPENAI = "OpenAI"
    GROQ = "Groq"
    ANTHROPIC = "Anthropic"
    GOOGLE = "Google"


provider_default_models: dict[LlmProvider, str] = {
    LlmProvider.OLLAMA: "llama3.1:8b",
    LlmProvider.OPENAI: "gpt-4o-mini",
    LlmProvider.GROQ: "llama3-70b-8192",
    LlmProvider.ANTHROPIC: "claude-3-5-sonnet-20240620",
    LlmProvider.GOOGLE: "gemini-1.5-pro-exp-0827",
}

provider_env_key_names: dict[LlmProvider, str] = {
    LlmProvider.OLLAMA: "",
    LlmProvider.OPENAI: "OPENAI_API_KEY",
    LlmProvider.GROQ: "GROQ_API_KEY",
    LlmProvider.ANTHROPIC: "ANTHROPIC_API_KEY",
    LlmProvider.GOOGLE: "GOOGLE_API_KEY",
}


llm_provider_types: list[LlmProvider] = list(LlmProvider)
provider_select_options: list[tuple[str, LlmProvider]] = [
    (
        p,
        LlmProvider(p),
    )
    for p in llm_provider_types
]


def get_llm_provider_from_str(llm_provider_str: str) -> LlmProvider:
    """Get the LLMProvider given a string matching one of its values."""
    for llm_provider in LlmProvider:
        if llm_provider.value == llm_provider_str:
            return llm_provider
    raise ValueError(f"Invalid LLM provider: {llm_provider_str}")



"""Output formatting utilities."""

from __future__ import annotations

import csv
import io
from pathlib import Path
from typing import Optional

from rich.syntax import Syntax
from rich.table import Table


def csv_to_table(data: str, title: str = "Results") -> Table:
    """Convert csv data to a table."""
    data = data.strip()
    table = Table(title=title)
    if not data:
        table.add_column("Empty", justify="left", style="cyan", no_wrap=True)
        return table
    reader = csv.DictReader(io.StringIO(data))
    if not reader.fieldnames:
        table.add_column("Empty", justify="left", style="cyan", no_wrap=True)
        return table
    data = list(reader)  # pyright: ignore
    for f in reader.fieldnames:
        table.add_column(f, justify="left", style="cyan", no_wrap=True)

    for row in data:
        table.add_row(*[v for n, v in row.items()])  # pyright: ignore
    return table


def csv_file_to_table(csv_file: Path, title: Optional[str] = None) -> Table:
    """Convert csv file to a table."""
    return csv_to_table(
        csv_file.read_text(encoding="utf-8").strip(),
        csv_file.name if title is None else title,
    )


def highlight_json(data: str) -> Syntax:
    """Highlight JSON data."""
    return Syntax(data, "json", background_color="default")


def highlight_json_file(json_file: Path) -> Syntax:
    """Highlight JSON data."""
    return highlight_json(json_file.read_text(encoding="utf-8").strip())



"""Enum for scraper choices."""

from enum import Enum


class CleanupType(str, Enum):
    """Enum for cleanup choices."""

    NONE = "none"
    BEFORE = "before"
    AFTER = "after"
    BOTH = "both"


class DisplayOutputFormat(str, Enum):
    """Enum for display output format choices."""

    MD = "md"
    CSV = "csv"
    JSON = "json"


class ScraperChoice(str, Enum):
    """Enum for scraper choices."""

    SELENIUM = "selenium"
    PLAYWRIGHT = "playwright"


class WaitType(str, Enum):
    """Enum for wait type choices."""

    NONE = "none"
    PAUSE = "pause"
    SLEEP = "sleep"
    IDLE = "idle"
    SELECTOR = "selector"
    TEXT = "text"



"""Main entry point for par_scrape."""

import csv
import json
import os
import shutil
import time
from datetime import datetime
from io import StringIO
from pathlib import Path
from typing import List, Optional, Annotated
from uuid import uuid4
from contextlib import nullcontext

import typer
from dotenv import load_dotenv
from openai import base_url
from rich.markdown import Markdown
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text

from par_scrape.enums import CleanupType, DisplayOutputFormat, ScraperChoice, WaitType
from par_scrape.scrape_data import (
    save_raw_data,
    create_dynamic_listing_model,
    create_listings_container_model,
    format_data,
    save_formatted_data,
)
from par_scrape.fetch_html import (
    fetch_html_selenium,
    fetch_html_playwright,
    html_to_markdown_with_readability,
)
from par_scrape.pricing import display_price_summary
from par_scrape.utils import console
from par_scrape import __version__, __application_title__
from par_scrape.lib.llm_providers import (
    LlmProvider,
    provider_default_models,
    provider_env_key_names,
)

# Load the .env file from the project folder
load_dotenv(dotenv_path=".env")
# Load the .env file from the users home folder
load_dotenv(dotenv_path=Path("~/.par-scrape.env").expanduser())

# Initialize Typer app
app = typer.Typer(help="Web scraping tool with options for Selenium or Playwright")


def version_callback(value: bool) -> None:
    """Print version and exit."""
    if value:
        print(f"{__application_title__}: {__version__}")
        raise typer.Exit()


# pylint: disable=too-many-statements,dangerous-default-value,too-many-arguments, too-many-locals, too-many-positional-arguments, too-many-branches
@app.command()
def main(
    url: Annotated[
        str, typer.Option("--url", "-u", help="URL to scrape")
    ] = "https://openai.com/api/pricing/",
    fields: Annotated[
        List[str],
        typer.Option("--fields", "-f", help="Fields to extract from the webpage"),
    ] = [
        "Model",
        "Pricing Input",
        "Pricing Output",
    ],
    scraper: Annotated[
        ScraperChoice,
        typer.Option(
            "--scraper",
            "-s",
            help="Scraper to use: 'selenium' or 'playwright'",
            case_sensitive=False,
        ),
    ] = ScraperChoice.PLAYWRIGHT,
    wait_type: Annotated[
        WaitType,
        typer.Option(
            "--wait-type",
            "-w",
            help="Method to use for page content load waiting",
            case_sensitive=False,
        ),
    ] = WaitType.SLEEP,
    wait_selector: Annotated[
        Optional[str],
        typer.Option(
            "--wait-selector",
            "-i",
            help="Selector or text to use for page content load waiting.",
        ),
    ] = None,
    headless: Annotated[
        bool,
        typer.Option("--headless", "-h", help="Run in headless mode (for Selenium)"),
    ] = False,
    sleep_time: Annotated[
        int,
        typer.Option(
            "--sleep-time", "-t", help="Time to sleep before scrolling (in seconds)"
        ),
    ] = 3,
    ai_provider: Annotated[
        LlmProvider,
        typer.Option("--ai-provider", "-a", help="AI provider to use for processing"),
    ] = LlmProvider.OPENAI,
    model: Annotated[
        Optional[str],
        typer.Option(
            "--model",
            "-m",
            help="AI model to use for processing. If not specified, a default model will be used.",
        ),
    ] = None,
    ai_base_url: Annotated[
        Optional[str],
        typer.Option(
            "--ai-base-url",
            "-b",
            help="Override the base URL for the AI provider.",
        ),
    ] = None,
    display_output: Annotated[
        Optional[DisplayOutputFormat],
        typer.Option(
            "--display-output",
            "-d",
            help="Display output in terminal (md, csv, or json)",
        ),
    ] = None,
    output_folder: Annotated[
        Path,
        typer.Option(
            "--output-folder", "-o", help="Specify the location of the output folder"
        ),
    ] = Path("./output"),
    silent: Annotated[
        bool,
        typer.Option("--silent", "-q", help="Run in silent mode, suppressing output"),
    ] = False,
    run_name: Annotated[
        str,
        typer.Option("--run-name", "-n", help="Specify a name for this run"),
    ] = "",
    version: Annotated[  # pylint: disable=unused-argument
        Optional[bool],
        typer.Option("--version", "-v", callback=version_callback, is_eager=True),
    ] = None,
    pricing: Annotated[
        bool,
        typer.Option("--pricing", help="Enable pricing summary display"),
    ] = False,
    cleanup: Annotated[
        CleanupType,
        typer.Option("--cleanup", "-c", help="How to handle cleanup of output folder."),
    ] = CleanupType.NONE,
    extraction_prompt: Annotated[
        Optional[Path],
        typer.Option(
            "--extraction-prompt", "-e", help="Path to the extraction prompt file"
        ),
    ] = None,
):
    """Scrape and analyze data from a website."""
    if not model:
        model = provider_default_models[ai_provider]

    if ai_provider != LlmProvider.OLLAMA:
        key_name = provider_env_key_names[ai_provider]
        if not os.environ.get(key_name):
            console.print(
                f"[bold red]{key_name} environment variable not set. Exiting...[/bold red]"
            )
            raise typer.Exit(1)
    with console.capture() if silent else nullcontext():
        if cleanup in [CleanupType.BEFORE, CleanupType.BOTH]:
            if os.path.exists(output_folder):
                shutil.rmtree(output_folder)
                console.print(
                    f"[bold green]Removed existing output folder: {output_folder}[/bold green]"
                )

        start_time = time.time()

        try:
            # Generate run_name if not provided
            if not run_name:
                run_name = datetime.now().strftime("%Y%m%d_%H%M%S")
            else:
                # Ensure run_name is filesystem-friendly
                run_name = "".join(
                    c for c in run_name if c.isalnum() or c in ("-", "_")
                )
                if not run_name:
                    run_name = str(uuid4())

            # Check if url is a local file
            is_local_file = os.path.isfile(url)
            source_type = "Local File" if is_local_file else "URL"

            # Display summary of options
            console.print(
                Panel.fit(
                    Text.assemble(
                        (f"{source_type}: ", "cyan"),
                        (f"{url}", "green"),
                        "\n",
                        ("AI Provider: ", "cyan"),
                        (f"{ai_provider.value}", "green"),
                        "\n",
                        ("Model: ", "cyan"),
                        (f"{model}", "green"),
                        "\n",
                        ("AI Provider Base URL: ", "cyan"),
                        (f"{base_url or 'default'}", "green"),
                        "\n",
                        ("Scraper: ", "cyan"),
                        (f"{scraper if not is_local_file else 'N/A'}", "green"),
                        "\n",
                        ("Headless: ", "cyan"),
                        (f"{headless if not is_local_file else 'N/A'}", "green"),
                        "\n",
                        ("Wait Type: ", "cyan"),
                        (f"{wait_type.value}", "green"),
                        "\n",
                        ("Wait Selector: ", "cyan"),
                        (
                            f"{wait_selector if wait_type in (WaitType.SELECTOR, WaitType.TEXT) else 'N/A'}",
                            "green",
                        ),
                        "\n",
                        ("Sleep Time: ", "cyan"),
                        (
                            f"{sleep_time if not is_local_file else 'N/A'} seconds",
                            "green",
                        ),
                        "\n",
                        ("Fields to extract: ", "cyan"),
                        (", ".join(fields), "green"),
                        "\n",
                        ("Display output: ", "cyan"),
                        (f"{display_output or 'None'}", "green"),
                        "\n",
                        ("Silent mode: ", "cyan"),
                        (f"{silent}", "green"),
                        "\n",
                        ("Cleanup: ", "cyan"),
                        (f"{cleanup}", "green"),
                    ),
                    title="[bold]Scraping Configuration",
                    border_style="bold",
                )
            )

            with console.status(
                "[bold green]Working on data extraction and processing..."
            ) as status:
                if is_local_file:
                    # Read local file
                    status.update("[bold cyan]Reading local file...")
                    with open(url, "rt", encoding="utf-8") as file:
                        markdown = file.read()
                    run_name = os.path.splitext(os.path.basename(url))[0].replace(
                        "rawData_", ""
                    )
                else:
                    # Scrape data
                    status.update("[bold cyan]Fetching HTML...")
                    if scraper == ScraperChoice.PLAYWRIGHT:
                        raw_html = fetch_html_playwright(
                            url, headless, wait_type, wait_selector, sleep_time
                        )
                    else:
                        raw_html = fetch_html_selenium(
                            url, headless, wait_type, wait_selector, sleep_time
                        )

                    status.update("[bold cyan]Converting HTML to Markdown...")
                    markdown = html_to_markdown_with_readability(raw_html)
                    # Save raw data
                    status.update("[bold cyan]Saving raw data...")
                    save_raw_data(markdown, run_name, output_folder)

                # Create the dynamic listing model
                status.update("[bold cyan]Creating dynamic models...")
                dynamic_listing_model = create_dynamic_listing_model(fields)
                dynamic_listings_container = create_listings_container_model(
                    dynamic_listing_model
                )

                # Format data
                status.update("[bold cyan]Formatting data...")
                formatted_data = format_data(
                    markdown,
                    dynamic_listings_container,
                    model,
                    ai_provider,
                    extraction_prompt,
                    ai_base_url,
                )
                if not formatted_data:
                    raise ValueError("No data was found by the scrape.")

                # Save formatted data
                status.update("[bold cyan]Saving formatted data...")
                _, file_paths = save_formatted_data(
                    formatted_data, run_name, output_folder
                )

                # Convert formatted_data back to text for token counting
                formatted_data_text = json.dumps(formatted_data.model_dump())

            # Display output if requested
            if display_output:
                if display_output.value in file_paths:
                    with open(
                        file_paths[display_output.value], "rt", encoding="utf-8"
                    ) as f:
                        content = f.read()
                    if display_output == DisplayOutputFormat.MD:
                        console.print(Markdown(content))
                    elif display_output == DisplayOutputFormat.CSV:
                        # Convert CSV to rich Table
                        table = Table(title="CSV Data")
                        csv_reader = csv.reader(StringIO(content))
                        headers = next(csv_reader)
                        for header in headers:
                            table.add_column(header, style="cyan")
                        for row in csv_reader:
                            table.add_row(*row)
                        console.print(table)
                    elif display_output == DisplayOutputFormat.JSON:
                        console.print(Syntax(content, "json"))
                else:
                    console.print(
                        f"[bold red]Invalid output type: {display_output.value}[/bold red]"
                    )

            duration = time.time() - start_time
            console.print(Panel.fit(f"Done in {duration:.2f} seconds."))
            # Display price summary
            if pricing:
                display_price_summary(status, model, markdown, formatted_data_text)

        except Exception as e:  # pylint: disable=broad-except
            # print(e)
            console.print(f"[bold red]An error occurred:[/bold red] {str(e)}")

        finally:
            if cleanup in [CleanupType.BOTH, CleanupType.AFTER]:
                with console.status("[bold yellow]Cleaning up..."):
                    if os.path.exists(output_folder):
                        shutil.rmtree(output_folder)
                        console.print(
                            f"[bold green]Removed output folder and its contents: {output_folder}[/bold green]"
                        )


if __name__ == "__main__":
    app()



"""Fetch HTML from URL and save it to a file."""

import logging
import os
import time
from typing import Optional

import html2text
from bs4 import BeautifulSoup
from playwright.sync_api import sync_playwright, expect
from selenium import webdriver
from selenium.webdriver.chrome.webdriver import WebDriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.common.exceptions import TimeoutException
from webdriver_manager.chrome import ChromeDriverManager

from .enums import WaitType
from .utils import console


def setup_selenium(headless: bool = True) -> WebDriver:
    """Set up ChromeDriver for Selenium."""
    logger = logging.getLogger("selenium")
    logger.setLevel(logging.DEBUG)

    options = Options()

    # adding arguments
    options.add_argument("--disable-gpu")
    options.add_argument("--disable-dev-shm-usage")
    options.add_argument("--window-size=1920,1080")
    options.add_argument("--window-position=-2400,-2400")

    options.add_experimental_option(
        "excludeSwitches", ["enable-logging"]
    )  # Disable logging
    options.add_argument("--silent")
    options.add_argument("--disable-extensions")

    # Enable headless mode if specified
    if headless:
        options.add_argument("--headless=new")

    # Randomize user-agent to mimic different users
    options.add_argument(
        "user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"  # pylint: disable=line-too-long
    )

    try:
        chromedriver_path = ChromeDriverManager().install()
        # console.log(chromedriver_path)

        service = Service(chromedriver_path, log_output=os.devnull)

        # Initialize the WebDriver
        driver = webdriver.Chrome(service=service, options=options)
        return driver
    except Exception as e:
        console.print(
            f"[bold red]Error initializing Chrome WebDriver:[/bold red] {str(e)}"
        )
        console.print(
            "[yellow]Please ensure Chrome is installed and up to date.[/yellow]"
        )
        console.print(
            "[yellow]Also, make sure the ChromeDriver version matches your Chrome browser version.[/yellow]"
        )
        raise


def fetch_html_selenium(
    url: str,
    headless: bool = True,
    wait_type: WaitType = WaitType.SLEEP,
    wait_selector: Optional[str] = None,
    sleep_time: int = 3,
) -> str:
    """
    Fetch HTML content from a URL using Selenium.

    Args:
        url (str): The URL to fetch HTML content from.
        headless (bool, optional): Whether to run the browser in headless mode. Defaults to True.
        wait_type (WaitType, optional): The type of wait to use. Defaults to WaitType.SLEEP.
        wait_selector (Optional[str], optional): The CSS selector to wait for. Defaults to None.
        sleep_time (int, optional): The time to sleep in seconds. Defaults to 3.

    Returns:
        str: The fetched HTML content as a string.
    """
    driver = setup_selenium(headless)
    driver.set_script_timeout(30)

    try:
        driver.get(url)
        try:
            # Wait for page to load
            if wait_type == WaitType.PAUSE:
                console.print("[yellow]Press Enter to continue...[/yellow]")
                input()
            elif wait_type == WaitType.SLEEP and sleep_time > 0:
                time.sleep(sleep_time)
            elif wait_type == WaitType.IDLE:
                time.sleep(1)
            elif wait_type == WaitType.SELECTOR:
                if wait_selector:
                    wait = WebDriverWait(driver, 10)
                    wait.until(
                        EC.presence_of_element_located((By.CSS_SELECTOR, wait_selector))
                    )
            elif wait_type == WaitType.TEXT:
                if wait_selector:
                    wait = WebDriverWait(driver, 10)
                    wait.until(
                        EC.text_to_be_present_in_element(
                            (By.TAG_NAME, "body"), wait_selector
                        )
                    )

            # Add more realistic actions like scrolling
            driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")
            time.sleep(3)  # Simulate time taken to scroll and read
        except TimeoutException:
            console.print("[yellow]Timed out waiting for condition.[/yellow]")

        return driver.page_source
    finally:
        driver.quit()


def fetch_html_playwright(
    url: str,
    headless: bool = True,
    wait_type: WaitType = WaitType.IDLE,
    wait_selector: Optional[str] = None,
    sleep_time: int = 3,
) -> str:
    """
    Fetch HTML content from a URL using Playwright.

    Args:
        url (str): The URL to fetch HTML from.
        headless (bool, optional): Whether to run the browser in headless mode. Defaults to True.
        wait_type (WaitType, optional): The type of wait to use. Defaults to WaitType.IDLE.
        wait_selector (str, optional): The CSS selector to wait for. Defaults to None.
        sleep_time (int, optional): The sleep time in seconds. Defaults to 3.
    Returns:
        str: The HTML content of the page.
    """
    with sync_playwright() as p:
        try:
            browser = p.chromium.launch(headless=headless)
        except Exception as e:  # pylint: disable=broad-except
            console.print(
                "[bold red]Error launching playwright browser:[/bold red] Make sure you install playwright: `uv tool install playwright` then run `playwright install chromium`."  # pylint: disable=line-too-long
                # pylint: disable=line-too-long
            )
            raise e
            # return ["" * len(urls)]
        context = browser.new_context(
            viewport={"width": 1280, "height": 1024},
            user_agent="Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36",  # pylint: disable=line-too-long
        )

        page = context.new_page()
        page.goto(url)

        if wait_type == WaitType.PAUSE:
            console.print("[yellow]Press Enter to continue...[/yellow]")
            input()
        elif wait_type == WaitType.SLEEP:
            # Add delays to mimic human behavior
            page.wait_for_timeout(sleep_time * 1000)  # Use the specified sleep time
        elif wait_type == WaitType.IDLE:
            page.wait_for_load_state("networkidle")  # domcontentloaded
        elif wait_type == WaitType.SELECTOR:
            if wait_selector:
                page.wait_for_selector(wait_selector)
            else:
                console.print(
                    "[bold yellow]Warning:[/bold yellow] Please specify a selector when using wait_type=selector."
                )
        elif wait_type == WaitType.TEXT:
            if wait_selector:
                expect(page.locator("body")).to_contain_text(wait_selector)
            else:
                console.print(
                    "[bold yellow]Warning:[/bold yellow] Please specify a selector when using wait_type=text."
                )

        # Add more realistic actions like scrolling
        page.evaluate("window.scrollTo(0, document.body.scrollHeight)")
        time.sleep(3)  # Simulate time taken to scroll and read

        html = page.content()
        browser.close()
        return html


def clean_html(html_content: str) -> str:
    """
    Clean HTML content by removing headers and footers.

    Args:
        html_content (str): The raw HTML content.

    Returns:
        str: The cleaned HTML content.
    """
    soup = BeautifulSoup(html_content, "html.parser")

    # Remove headers and footers based on common HTML tags or classes
    for element in soup.find_all(["header", "footer"]):
        element.decompose()  # Remove these tags and their content

    return str(soup)


def html_to_markdown_with_readability(html_content: str) -> str:
    """
    Convert HTML content to Markdown format.

    Args:
        html_content (str): The HTML content to convert.

    Returns:
        str: The converted markdown content.
    """
    cleaned_html = clean_html(html_content)

    # Convert to markdown
    markdown_converter = html2text.HTML2Text()
    markdown_converter.ignore_links = False
    markdown_content = markdown_converter.handle(cleaned_html)

    return markdown_content



"""Models for rag related tasks."""

from __future__ import annotations

import os
import warnings
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from langchain._api import LangChainDeprecationWarning
from langchain_anthropic import ChatAnthropic
from langchain_community.embeddings import OllamaEmbeddings
from langchain_community.llms.ollama import Ollama
from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseChatModel
from langchain_core.language_models import BaseLanguageModel
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain_google_genai import GoogleGenerativeAI
from langchain_groq import ChatGroq
from langchain_ollama import ChatOllama
from langchain_openai import ChatOpenAI
from langchain_openai import OpenAI
from langchain_openai import OpenAIEmbeddings


from par_scrape.lib.llm_providers import LlmProvider


warnings.simplefilter("ignore", category=LangChainDeprecationWarning)

ollama_host = os.environ.get("OLLAMA_HOST", "http://localhost:11434")


class LlmMode(str, Enum):
    """LLM mode types."""

    BASE = "Base"
    CHAT = "Chat"
    EMBEDDINGS = "Embeddings"


llm_modes: list[LlmMode] = list(LlmMode)


@dataclass
class LlmConfig:
    """Configuration for Llm."""

    provider: LlmProvider
    model_name: str
    mode: LlmMode = LlmMode.CHAT
    temperature: float = 0.5
    streaming: bool = False
    base_url: Optional[str] = None
    class_name: str = "LlmConfig"

    def to_json(self) -> dict:
        """Return dict for use with json"""
        return {
            "class_name": self.__class__.__name__,
            "provider": self.provider,
            "model_name": self.model_name,
            "mode": self.mode,
            "temperature": self.temperature,
            "streaming": self.streaming,
            "base_url": self.base_url,
        }

    @staticmethod
    def from_json(data: dict) -> LlmConfig:
        """Create instance from json data"""
        if data["class_name"] != "LlmConfig":
            raise ValueError(f"Invalid config class: {data['class_name']}")
        del data["class_name"]
        return LlmConfig(**data)

    def clone(self) -> LlmConfig:
        """Create a clone of the LlmConfig."""
        return LlmConfig(
            provider=self.provider,
            model_name=self.model_name,
            mode=self.mode,
            temperature=self.temperature,
            streaming=self.streaming,
            base_url=self.base_url,
        )

    # pylint: disable=too-many-return-statements,too-many-branches
    def _build_llm(self) -> BaseLanguageModel | BaseChatModel | Embeddings:
        """Build the LLM."""
        if self.provider == LlmProvider.OLLAMA:
            if self.mode == LlmMode.BASE:
                return Ollama(
                    model=self.model_name,
                    temperature=self.temperature,
                    base_url=self.base_url or ollama_host,
                )
            if self.mode == LlmMode.CHAT:
                return ChatOllama(
                    model=self.model_name,
                    temperature=self.temperature,
                    base_url=self.base_url or ollama_host,
                )
            if self.mode == LlmMode.EMBEDDINGS:
                return OllamaEmbeddings(
                    base_url=self.base_url or ollama_host, model=self.model_name
                )
        elif self.provider == LlmProvider.OPENAI:
            if self.mode == LlmMode.BASE:
                return OpenAI(
                    model=self.model_name,
                    temperature=self.temperature,
                    streaming=self.streaming,
                    base_url=self.base_url,
                )
            if self.mode == LlmMode.CHAT:
                return ChatOpenAI(
                    model=self.model_name,
                    temperature=self.temperature,
                    streaming=self.streaming,
                    base_url=self.base_url,
                )
            if self.mode == LlmMode.EMBEDDINGS:
                return OpenAIEmbeddings(model=self.model_name)
        elif self.provider == LlmProvider.GROQ:
            if self.mode == LlmMode.BASE:
                raise ValueError(
                    f"{self.provider} provider does not support mode {self.mode}"
                )
            if self.mode == LlmMode.CHAT:
                return ChatGroq(
                    model=self.model_name,
                    temperature=self.temperature,
                    streaming=self.streaming,
                    base_url=self.base_url,
                )  # pyright: ignore [reportCallIssue]
            if self.mode == LlmMode.EMBEDDINGS:
                raise ValueError(
                    f"{self.provider} provider does not support mode {self.mode}"
                )
        elif self.provider == LlmProvider.ANTHROPIC:
            if self.mode == LlmMode.BASE:
                raise ValueError(
                    f"{self.provider} provider does not support mode {self.mode}"
                )
            if self.mode == LlmMode.CHAT:
                return ChatAnthropic(  # pyright: ignore [reportCallIssue]
                    model=self.model_name,  # pyright: ignore [reportCallIssue]
                    temperature=self.temperature,
                    streaming=self.streaming,
                    default_headers={"anthropic-beta": "tools-2024-05-16"},
                    base_url=self.base_url,
                    # max_tokens=8_192,
                )
            if self.mode == LlmMode.EMBEDDINGS:
                raise ValueError(
                    f"{self.provider} provider does not support mode {self.mode}"
                )
        elif self.provider == LlmProvider.GOOGLE:
            if self.mode == LlmMode.BASE:
                return GoogleGenerativeAI(
                    model=self.model_name, temperature=self.temperature
                )
            if self.mode == LlmMode.CHAT:
                return ChatGoogleGenerativeAI(
                    model=self.model_name, temperature=self.temperature
                )
            if self.mode == LlmMode.EMBEDDINGS:
                raise ValueError(
                    f"{self.provider} provider does not support mode {self.mode}"
                )
        raise ValueError(
            f"Invalid LLM provider '{self.provider}' or mode '{self.mode}'"
        )

    def build_llm_model(self) -> BaseLanguageModel:
        """Build the LLM model."""
        llm = self._build_llm()
        if isinstance(llm, BaseLanguageModel):
            return llm
        raise ValueError(f"LLM provider '{self.provider}' does not support base mode.")

    def build_chat_model(self) -> BaseChatModel:
        """Build the chat model."""
        llm = self._build_llm()
        if isinstance(llm, BaseChatModel):
            return llm
        raise ValueError(f"LLM provider '{self.provider}' does not support chat mode.")

    def build_embeddings(self) -> Embeddings:
        """Build the embeddings."""
        llm = self._build_llm()
        if isinstance(llm, Embeddings):
            return llm
        raise ValueError(f"LLM mode '{self.mode}' does not support embeddings.")



"""Pricing functions for Par Scrape."""

from typing import Tuple
import tiktoken
from rich.panel import Panel
from rich.status import Status
from rich.text import Text

from par_scrape.utils import console, estimate_tokens

pricing = {
    "gpt-4o": {
        "input": 0.000005,  # $5.00 per 1M input tokens
        "output": 0.000015,  # $15.00 per 1M output tokens
    },
    "gpt-4o-2024-08-06": {
        "input": 0.0000025,  # $2.50 per 1M input tokens
        "output": 0.00001,  # $10.00 per 1M output tokens
    },
    "gpt-4o-2024-05-13": {
        "input": 0.000005,  # $5.00 per 1M input tokens
        "output": 0.000015,  # $15.00 per 1M output tokens
    },
    "gpt-4o-mini": {
        "input": 0.00000015,  # $0.15 per 1M input tokens
        "output": 0.0000006,  # $0.60 per 1M output tokens
    },
    "gpt-4o-mini-2024-07-18": {
        "input": 0.00000015,  # $0.15 per 1M input tokens
        "output": 0.0000006,  # $0.60 per 1M output tokens
    },
    "gpt-4": {
        "input": 0.00003,  # $30.00 per 1M input tokens
        "output": 0.00006,  # $60.00 per 1M output tokens
    },
    "gpt-4-32k": {
        "input": 0.00006,  # $60.00 per 1M input tokens
        "output": 0.00012,  # $120.00 per 1M output tokens
    },
    "gpt-4-turbo": {
        "input": 0.00001,  # $10.00 per 1M input tokens
        "output": 0.00003,  # $30.00 per 1M output tokens
    },
    "gpt-4-turbo-2024-04-09": {
        "input": 0.00001,  # $10.00 per 1M input tokens
        "output": 0.00003,  # $30.00 per 1M output tokens
    },
    "gpt-3.5-turbo-0125": {
        "input": 0.0000005,  # $0.50 per 1M input tokens
        "output": 0.0000015,  # $1.50 per 1M output tokens
    },
    "claude-3-5-sonnet-20240620": {
        "input": 0.000003,  # $3.0 per 1M input tokens
        "output": 0.000015,  # $15.0 per 1M output tokens
    },
    "claude-3-haiku-20240307": {
        "input": 0.00000025,  # $0.25 per 1M input tokens
        "output": 0.00000125,  # $1.25 per 1M output tokens
    },
    "claude-3-sonnet-20240229": {
        "input": 0.000003,  # $3.0 per 1M input tokens
        "output": 0.000015,  # $15.0 per 1M output tokens
    },
    "claude-3-opus-20240229": {
        "input": 0.000015,  # $15.0 per 1M input tokens
        "output": 0.000075,  # $75.0 per 1M output tokens
    },
}


def calculate_price(
    input_text: str, output_text: str, model: str
) -> Tuple[int, int, float]:
    """
    Calculate the price for processing input and output text using a specified model.

    Args:
        input_text (str): The input text.
        output_text (str): The output text.
        model (str, optional): The model to use for pricing. Defaults to model_used.

    Returns:
        Tuple[int, int, float]: A tuple containing the input token count, output token count, and total cost.
        Returns (0, 0, 0.0) if there's an error during calculation.
    """
    input_token_count = 0
    output_token_count = 0
    total_cost = 0.0
    try:
        try:
            # Initialize the encoder for the specific model
            encoder = tiktoken.encoding_for_model(model)

            # Encode the input text to get the number of input tokens
            input_token_count = len(encoder.encode(input_text))

            # Encode the output text to get the number of output tokens
            output_token_count = len(encoder.encode(output_text))
        except Exception as _:  # pylint: disable=broad-except
            console.print(
                Panel.fit(
                    Text.assemble(
                        ("Could not get encoder. Falling back to estimator.", "yellow")
                    )
                )
            )
            input_token_count = estimate_tokens(input_text)
            output_token_count = estimate_tokens(output_text)

        # Calculate the costs
        if model in pricing:
            pricing_in = pricing[model]["input"]
            pricing_out = pricing[model]["output"]
        else:
            pricing_in = 0
            pricing_out = 0

        input_cost = input_token_count * pricing_in
        output_cost = output_token_count * pricing_out
        total_cost = input_cost + output_cost
    except Exception as _:  # pylint: disable=broad-except
        console.print(
            Panel.fit(
                Text.assemble(("Error calculating token usage and or cost.", "red"))
            )
        )
    return input_token_count, output_token_count, total_cost


def display_price_summary(
    status: Status, model: str, markdown, formatted_data_text
) -> None:
    """Display the price summary."""
    status.update("[bold cyan]Calculating token usage and cost...")
    input_tokens, output_tokens, total_cost = calculate_price(
        markdown, formatted_data_text, model=model
    )
    if input_tokens == 0 or output_tokens == 0:
        console.print(
            Panel.fit(
                Text.assemble(("Could not calculate token usage and cost.", "yellow"))
            )
        )
    else:
        console.print(
            Panel.fit(
                Text.assemble(
                    ("Input token count: ", "cyan"),
                    (f"{input_tokens}", "green"),
                    "\n",
                    ("Output token count: ", "cyan"),
                    (f"{output_tokens}", "green"),
                    "\n",
                    ("Estimated total cost: ", "cyan"),
                    (f"${total_cost:.4f}", "green bold"),
                ),
                title="[bold]Summary",
                border_style="bold",
            )
        )
